import { compileTemplate, compileStyleAsync, parse } from '@vue/compiler-sfc';
import createDebugger from 'debug';
import hash from 'hash-sum';
import { relative, basename } from 'path';
import qs from 'querystring';
import { createFilter } from 'rollup-pluginutils';
import { encode } from 'sourcemap-codec';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function genCSSModulesCode( // @ts-ignore
id, index, request, moduleName) {
  var styleVar = "style" + index;
  var code = "\nimport " + styleVar + " from " + request; // inject variable

  var name = typeof moduleName === 'string' ? moduleName : '$style';
  code += "\ncssModules[\"" + name + "\"] = " + styleVar;
  return code;
}

try {
  require.resolve('@vue/compiler-sfc');
} catch (e) {
  throw new Error('vue-loader requires @vue/compiler-sfc to be present in the dependency ' + 'tree.');
}
var debug =
/*#__PURE__*/
createDebugger('rollup-plugin-vue');
var defaultOptions = {
  include: /\.vue$/,
  exclude: [],
  target: 'browser',
  exposeFilename: false
};
function PluginVue(userOptions) {
  if (userOptions === void 0) {
    userOptions = {};
  }

  var options = _extends({}, defaultOptions, {}, userOptions);

  var isServer = options.target === 'node';
  var isProduction = process.env.NODE_ENV === 'production' || process.env.BUILD === 'production';
  var rootContext = process.cwd();
  var filter = createFilter(options.include, options.exclude);
  return {
    name: 'vue',
    resolveId: function resolveId(id) {
      var query = parseVuePartRequest(id);

      if (query.vue) {
        debug("resolveId(" + id + ")");
        return id;
      }

      return undefined;
    },
    load: function load(id) {
      var query = parseVuePartRequest(id);

      if (query.vue) {
        var descriptor = getDescriptor(query.filename);
        var block = query.type === 'template' ? descriptor.template : query.type === 'script' ? descriptor.script : query.type === 'style' ? descriptor.styles[query.index] : query.type === 'custom' ? descriptor.customBlocks[query.index] : null;

        if (block) {
          var result = {
            code: block.content,
            map: normalizeSourceMap(block.map)
          };

          if (query.type === 'template') {
            // generate source mapping for each character.
            result.map.mappings = encode(result.code.split(/\r?\n/).map(function (line, index) {
              var segments = [];

              for (var i = 0; i < line.length; ++i) {
                segments.push([i, 0, block.loc.start.line + index - 1, i]);
              }

              return segments;
            }));
          }

          debug("load(" + id + ")", '\n' + result.code + '\n//# sourceMappingURL=data:application/json;charset=utf-8;base64,' + Buffer.from(JSON.stringify(result.map), 'utf-8').toString('base64'));
          return result;
        }
      }

      return undefined;
    },
    transform: function transform(code, id) {
      try {
        var _exit2 = false;

        var _this2 = this;

        var query = parseVuePartRequest(id);

        if (query.vue) {
          var descriptor = getDescriptor(query.filename);

          var _temp2 = function () {
            if (query.type === 'template') {
              debug("transform(" + id + ")");
              var block = descriptor.template;
              var result = compileTemplate({
                filename: query.filename,
                source: code,
                preprocessLang: block.lang,
                compiler: options.compiler,
                compilerOptions: options.compilerOptions,
                transformAssetUrls: options.transformAssetUrls
              });

              if (result.errors.length) {
                result.errors.forEach(function (error) {
                  return _this2.error(typeof error === 'string' ? {
                    id: query.filename,
                    message: error
                  } : createRollupError(query.filename, error));
                });
                _exit2 = true;
                return null;
              }

              if (result.tips.length) {
                result.tips.forEach(function (tip) {
                  return _this2.warn({
                    id: query.filename,
                    message: tip
                  });
                });
              }

              _exit2 = true;
              return {
                code: result.code,
                map: normalizeSourceMap(result.map)
              };
            } else return function () {
              if (query.type === 'style' && query.scoped) {
                debug("transform(" + id + ")");
                var _block = descriptor.styles[query.index];
                return Promise.resolve(compileStyleAsync({
                  filename: query.filename,
                  id: "data-v-" + query.id,
                  source: _block.content,
                  scoped: query.scoped,
                  trim: true
                })).then(function (result) {
                  if (result.errors.length) {
                    result.errors.forEach(function (error) {
                      return _this2.error({
                        id: query.filename,
                        message: error.message
                      });
                    });
                    _exit2 = true;
                    return null;
                  }

                  _exit2 = true;
                  return {
                    code: result.code,
                    map: normalizeSourceMap(result.map)
                  };
                });
              }
            }();
          }();

          return Promise.resolve(_temp2 && _temp2.then ? _temp2.then(function (_result2) {
            return _exit2 ? _result2 : null;
          }) : _exit2 ? _temp2 : null);
        } else if (filter(id)) {
          debug("transform(" + id + ")");

          var _parseSFC = parseSFC(code, id, rootContext),
              _descriptor = _parseSFC.descriptor,
              errors = _parseSFC.errors;

          if (errors.length) {
            errors.forEach(function (error) {
              return _this2.error(createRollupError(id, error));
            });
            return Promise.resolve(null);
          } // module id for scoped CSS & hot-reload


          var output = transformVueSFC(code, id, _descriptor, {
            rootContext: rootContext,
            isProduction: isProduction,
            isServer: isServer
          }, options);
          debug('transient .vue file:', '\n' + output + '\n');
          return Promise.resolve({
            code: output,
            map: {
              mappings: ''
            }
          });
        } else {
          return Promise.resolve(null);
        }
      } catch (e) {
        return Promise.reject(e);
      }
    }
  };
}

function parseVuePartRequest(id) {
  var _id$split = id.split('?', 2),
      filename = _id$split[0],
      query = _id$split[1];

  if (!query) return {
    vue: false,
    filename: filename
  };
  var raw = qs.parse(query);

  if ('vue' in raw) {
    return _extends({}, raw, {
      filename: filename,
      vue: true,
      type: raw.type,
      index: Number(raw.index),
      scoped: 'scoped' in raw,
      module: raw.module
    });
  }

  return {
    vue: false,
    filename: filename
  };
}

var cache =
/*#__PURE__*/
new Map();

function getDescriptor(id) {
  if (cache.has(id)) {
    return cache.get(id);
  }

  throw new Error(id + " is not parsed it yet");
}

function parseSFC(code, id, sourceRoot) {
  var _parse = parse(code, {
    sourceMap: true,
    filename: id,
    sourceRoot: sourceRoot,
    pad: 'line'
  }),
      descriptor = _parse.descriptor,
      errors = _parse.errors;

  cache.set(id, descriptor);
  return {
    descriptor: descriptor,
    errors: errors
  };
}

function transformVueSFC(code, resourcePath, descriptor, _ref, options) {
  var rootContext = _ref.rootContext,
      isProduction = _ref.isProduction;
  var shortFilePath = relative(rootContext, resourcePath).replace(/^(\.\.[\/\\])+/, '').replace(/\\/g, '/');
  var id = hash(isProduction ? shortFilePath + '\n' + code : shortFilePath); // feature information

  var hasScoped = descriptor.styles.some(function (s) {
    return s.scoped;
  });
  var templateImport = getTemplateCode(descriptor, resourcePath, id, hasScoped);
  var scriptImport = getScriptCode(descriptor, resourcePath);
  var stylesCode = getStyleCode(descriptor, resourcePath, id);
  var output = [scriptImport, templateImport, stylesCode, "script.render = render"];

  if (hasScoped) {
    output.push("script.__scopeId = " + _("data-v-" + id));
  }

  if (!isProduction) {
    output.push("script.__file = " + _(shortFilePath));
  } else if (options.exposeFilename) {
    output.push("script.__file = " + _(basename(shortFilePath)));
  }

  output.push('export default script');
  return output.join('\n');
}

function getTemplateCode(descriptor, resourcePath, id, hasScoped) {
  var templateImport = "const render = () => {}";
  var templateRequest;

  if (descriptor.template) {
    var src = descriptor.template.src || resourcePath;
    var idQuery = "&id=" + id;
    var scopedQuery = hasScoped ? "&scoped=true" : "";
    var attrsQuery = attrsToQuery(descriptor.template.attrs);
    var query = "?vue&type=template" + idQuery + scopedQuery + attrsQuery;
    templateRequest = _(src + query);
    templateImport = "import { render } from " + templateRequest;
  }

  return templateImport;
}

function getScriptCode(descriptor, resourcePath) {
  var scriptImport = "const script = {}";

  if (descriptor.script) {
    var src = descriptor.script.src || resourcePath;
    var attrsQuery = attrsToQuery(descriptor.script.attrs, 'js');
    var query = "?vue&type=script" + attrsQuery;

    var scriptRequest = _(src + query);

    scriptImport = "import script from " + scriptRequest + "\n" + ("export * from " + scriptRequest); // support named exports
  }

  return scriptImport;
}

function getStyleCode(descriptor, resourcePath, id) {
  var stylesCode = "";
  var hasCSSModules = false;

  if (descriptor.styles.length) {
    descriptor.styles.forEach(function (style, i) {
      var src = style.src || resourcePath;
      var attrsQuery = attrsToQuery(style.attrs, 'css'); // make sure to only pass id when necessary so that we don't inject
      // duplicate tags when multiple components import the same css file

      var idQuery = style.scoped ? "&id=" + id : "";
      var query = "?vue&type=style&index=" + i + idQuery + attrsQuery;

      var styleRequest = _(src + query);

      if (style.module) {
        if (!hasCSSModules) {
          stylesCode += "const cssModules = script.__cssModules = {}";
          hasCSSModules = true;
        }

        stylesCode += genCSSModulesCode(id, i, styleRequest, style.module);
      } else {
        stylesCode += "\nimport " + styleRequest;
      } // TODO SSR critical CSS collection

    });
  }

  return stylesCode;
}

function createRollupError(id, error) {
  return {
    id: id,
    plugin: 'vue',
    pluginCode: String(error.code),
    message: error.message,
    frame: error.loc.source,
    parserError: error,
    loc: error.loc ? {
      file: id,
      line: error.loc.start.line,
      column: error.loc.start.column
    } : undefined
  };
} // these are built-in query parameters so should be ignored
// if the user happen to add them as attrs


var ignoreList = ['id', 'index', 'src', 'type'];

function attrsToQuery(attrs, langFallback) {
  var query = "";

  for (var name in attrs) {
    var value = attrs[name];

    if (!ignoreList.includes(name)) {
      query += "&" + qs.escape(name) + "=" + (value ? qs.escape(String(value)) : "");
    }
  }

  if (langFallback && !("lang" in attrs)) {
    query += "&lang." + langFallback;
  }

  return query;
}

function _(any) {
  return JSON.stringify(any);
}

function normalizeSourceMap(map) {
  if (!map) return null;
  return {
    version: Number(map.version),
    file: map.file,
    names: map.names,
    sources: map.sources,
    sourceRoot: map.sourceRoot,
    sourcesContent: map.sourcesContent,
    mappings: typeof map.mappings === 'string' ? map.mappings : ''
  };
}

export default PluginVue;
//# sourceMappingURL=rollup-plugin-vue.esm.js.map
